from dataclasses import dataclass
from enum import Enum, auto
from pathlib import PurePosixPath, PurePath
from typing import Dict, Tuple, Callable, List

from aim_build.typedefs import StringList
from aim_build.utils import prepend_paths, to_native_path, to_pure_posix_path, relpaths


class BuildTypes(Enum):
    staticLibrary = auto()
    dynamicLibrary = auto()
    executable = auto()
    headerOnly = auto()
    libraryReference = auto()


@dataclass
class LibraryInformation:
    name: str
    path: str
    type: str


def get_project_dir(build: Dict, target_file: Dict):
    root_dir = target_file["projectRoot"]
    project_dir = build["build_dir"] / root_dir
    return project_dir


def find_build(build_name: str, builds: Dict) -> Dict:
    # Note, this should never fail, as required dependencies are checked by the schema.
    for build in builds:
        if build["name"] == build_name:
            return build

    raise RuntimeError(f"Failed to find build with name: {build_name}")


def find_builds_of_type(build_type: str, builds: Dict) -> List[Dict]:
    return [build for build in builds if build["buildRule"] == build_type]


def get_include_paths(include_paths: List[PurePath], build_dir: PurePath) -> List[PurePath]:
    include_paths = [to_native_path(path) for path in include_paths]
    abs_paths = [PurePosixPath(str(p).replace("\\", "/")) for p in include_paths if p.is_absolute()]
    rel_paths = [PurePosixPath(str(p).replace("\\", "/")) for p in include_paths if not p.is_absolute()]
    rel_paths = prepend_paths(build_dir, rel_paths)

    includes = abs_paths + rel_paths
    return includes


def get_toolchain_and_flags(build: Dict,
                            target_file: Dict) \
        -> Tuple[str, str, StringList, StringList, str, StringList]:
    local_compiler = build.get("compiler", None)
    local_archiver = build.get("archiver", None)
    local_flags = build.get("flags", None)
    local_defines = build.get("defines", None)
    local_linker = build.get("linker", None)
    local_linker_flags = build.get("linker_flags", None)

    compiler = local_compiler if local_compiler else target_file["compiler"]
    archiver = local_archiver if local_archiver else target_file["archiver"]
    cxx_flags = local_flags if local_flags else target_file.get("flags", [])
    defines = local_defines if local_defines else target_file.get("defines", [])
    global_linker = target_file.get("linker", compiler)

    linker = local_compiler if local_linker else global_linker
    linker_flags = local_linker_flags if local_linker_flags else cxx_flags

    return compiler, archiver, cxx_flags, defines, linker, linker_flags


def get_src_files(build: Dict, target_file: Dict) -> StringList:
    project_dir = get_project_dir(build, target_file)
    paths = build["sourceFiles"]
    paths = [to_native_path(path) for path in paths]

    # Resolve relative paths to the build directory, and leave absolute paths alone.
    #
    abs_paths = [path for path in paths if path.is_absolute() is True]
    rel_paths = [path for path in paths if path.is_absolute() is False]

    src_paths = []
    for path in abs_paths:
        if path.stem == "*":
            globbed_files = list(path.parent.glob(path.name))
            src_paths += globbed_files
        else:
            src_paths.append(path)

    build_path = build["build_dir"]
    for path in rel_paths:
        path = project_dir / path
        if path.stem == "*":
            globbed_files = list(path.parent.glob(path.name))

            rel_paths = relpaths(globbed_files, build_path)
            src_paths += rel_paths
        else:
            rel_path = relpaths(path, build_path)
            src_paths += rel_path

    src_paths = [to_pure_posix_path(path) for path in src_paths]
    return [str(file) for file in src_paths]


def get_required_library_information(build: Dict,
                                     parsed_toml: Dict) -> List[LibraryInformation]:
    requires = build.get("requires", [])
    if not requires:
        return []

    build_names = []  # Used to prevent duplicates.
    result = []

    library_types = [BuildTypes.staticLibrary, BuildTypes.dynamicLibrary]

    for required in requires:
        the_dep = find_build(required, parsed_toml["builds"])
        build_type = BuildTypes[the_dep["buildRule"]]
        if build_type not in library_types:
            continue

        build_name = the_dep["name"]
        if build_name not in build_names:
            build_names.append(build_name)

            # If we are going to dynamically load the library, then we don't want to generate linker flags for it.
            dynamic_loading = the_dep.get("dynamicLoading", False)
            if dynamic_loading:
                continue

            lib_info = LibraryInformation(
                the_dep["outputName"],
                the_dep["name"],
                the_dep["buildRule"]
            )
            result.append(lib_info)

    return result


def get_reference_library_information(
        build: Dict, parsed_toml: Dict
) -> Tuple[List[str], List[str]]:
    requires = build.get("requires", [])
    if not requires:
        return [], []

    build_names = []  # Used to prevent duplicates.
    libraries = []
    library_paths = []
    for required in requires:
        the_dep = find_build(required, parsed_toml["builds"])
        build_type = BuildTypes[the_dep["buildRule"]]
        if build_type != BuildTypes.libraryReference:
            continue

        build_name = the_dep["name"]
        if build_name not in build_names:
            build_names.append(build_name)
            libraries += the_dep.get("libraries", [])
            library_paths += the_dep.get("libraryPaths", [])

    return libraries, library_paths


def get_full_library_name_convention(
        lib_infos: List[LibraryInformation],
        static_convention_func: Callable[[str], str],
        dynamic_convention_func: Callable[[str], str],
) -> StringList:
    # Here we just need to manage the fact that the linker's library flag (-l) needs the library name without
    # lib{name}.a/.so but the build dependency rule does need the full convention to find the build rule in the
    # build.ninja file.
    full_library_names = []
    for info in lib_infos:
        build_type = BuildTypes[info.type]
        if build_type == BuildTypes.staticLibrary:
            full_library_names.append(static_convention_func(info.name))
        elif build_type == BuildTypes.dynamicLibrary:
            full_library_names.append(dynamic_convention_func(info.name))

    return full_library_names



import functools
from pathlib import PureWindowsPath, PurePosixPath
from typing import Callable
from typing import Dict, Tuple, List, Union

from ninja_syntax import Writer

from aim_build import commonbuilds
from aim_build.commonbuilds import BuildTypes, LibraryInformation
from aim_build.typedefs import StringList, PathList
from aim_build.utils import prefix, postfix, src_to_obj, prepend_paths, to_str, to_native_path, wrap_quotes

USING_RELATIVE_OUTPUTS = False

PrefixIncludePath = functools.partial(prefix, "/I")
PrefixLibraryPath = functools.partial(prefix, "/LIBPATH:")
PrefixLibrary = functools.partial(prefix, "")
PrefixHashDefine = functools.partial(prefix, "/D")
PostFixLib = functools.partial(postfix, ".lib")
ToObjectFiles = src_to_obj


def add_compile(nfw):
    # Note, there cannot be a space between /Fo and $out
    command = "$compiler $flags $defines $includes /showIncludes /c $in /Fo$out"
    nfw.rule(
        name="compile",
        description="Compile source files to object files",
        deps="msvc",
        depfile="deps.d",
        command=command,
    )
    nfw.newline()


def add_ar(nfw):
    nfw.rule(
        name="archive",
        description="Combine object files into an archive",
        command="llvm-ar cr $out $in",
    )
    nfw.newline()


def add_exe(nfw):
    command = (
        "$compiler $flags $defines $includes $in /link /out:$exe_name $linker_args"
    )
    nfw.rule(name="exe", description="Build an executable.", command=command)
    nfw.newline()


def add_shared(nfw):
    command = "$compiler $flags $defines $includes $in /link /DLL /out:$lib_name $linker_args"
    nfw.rule(name="shared", description="Build a shared library.", command=command)
    nfw.newline()


# TODO: Should take version strings as well?
def windows_add_static_library_naming_convention(library_name: str) -> str:
    return f"lib{library_name}.lib"


def windows_add_dynamic_library_naming_convention(library_name: str) -> str:
    return f"lib{library_name}.dll"


def convert_posix_to_windows(paths: Union[StringList, List[PurePosixPath]]):
    return [str(path).replace("/", "\\") for path in paths]


def windows_convert_strings_to_paths(paths: StringList):
    return [PureWindowsPath(path) for path in convert_posix_to_windows(paths)]


def convert_to_implicit_library_files(library):
    return "lib" + library + ".exp", "lib" + library + ".lib"


def get_external_libraries_paths(build):
    directory = PureWindowsPath(build["directory"])
    library_paths = build.get("libraryPaths", [])
    library_paths = convert_posix_to_windows(library_paths)
    library_paths = prepend_paths(directory, library_paths)
    return library_paths


def get_external_libraries_names(build: Dict) -> Tuple[StringList, StringList]:
    libraries = build.get("libraries", [])
    link_libraries = PrefixLibrary(libraries)
    return libraries, link_libraries


def get_external_libraries_information(build: Dict) -> Tuple[StringList, PathList]:
    libraries, _ = get_external_libraries_names(build)
    library_paths = get_external_libraries_paths(build)
    return libraries, library_paths


def get_library_information(lib_infos: List[LibraryInformation]) -> Tuple[List, List, List, List]:
    exps = []
    libs = []
    # This is a bit confusing, we need just the library names and exp file names as this is what the linker needs, but
    # ninja needs the full path to the exp file.
    implicit_exps = []
    implicit_libs = []

    for info in lib_infos:
        if info.type == "dynamicLib":
            exp, lib = convert_to_implicit_library_files(info.name)
            exps.append(exp)
            libs.append(lib)
            if USING_RELATIVE_OUTPUTS:
                implicit_exps.append(str(PureWindowsPath(info.path) / exp))
                implicit_libs.append(str(PureWindowsPath(info.path) / lib))
            else:
                implicit_exps.append(exp)
                implicit_libs.append(lib)
        else:
            lib = "lib" + info.name + ".lib"
            libs.append(lib)
            if USING_RELATIVE_OUTPUTS:
                implicit_libs.append(str(PureWindowsPath(info.path) / lib))
            else:
                implicit_libs.append(lib)

    return exps, libs, implicit_exps, implicit_libs


def get_includes_for_build(build: Dict, parsed_toml: Dict) -> StringList:
    requires = [build["name"]] + build.get("requires", [])

    include_paths = set()

    project_root = PureWindowsPath(parsed_toml["projectRoot"])

    for required in requires:
        the_dep = commonbuilds.find_build(required, parsed_toml["builds"])

        includes = the_dep.get("includePaths", [])
        includes = windows_convert_strings_to_paths(includes)
        includes = commonbuilds.get_include_paths(includes, project_root)
        include_paths.update(includes)

    include_paths = [to_native_path(path) for path in include_paths]
    include_paths = [str(path) for path in include_paths]

    include_paths.sort()
    include_paths = wrap_quotes(include_paths)
    include_args = PrefixIncludePath(include_paths)

    return include_args


def get_src_for_build(build: Dict, parsed_toml: Dict) -> List[PureWindowsPath]:
    files = commonbuilds.get_src_files(build, parsed_toml)
    return windows_convert_strings_to_paths(files)


def add_compile_rule(writer: Writer,
                     build: Dict,
                     target_file: Dict,
                     includes: StringList,
                     extra_flags: StringList = None):
    build_name = build["name"]

    compiler, _, cxx_flags, defines, __, ___ = commonbuilds.get_toolchain_and_flags(
        build, target_file
    )
    defines = PrefixHashDefine(defines)

    if extra_flags:
        cxx_flags = extra_flags + cxx_flags

    src_files = get_src_for_build(build, target_file)
    obj_files = src_to_obj(src_files)
    obj_files = prepend_paths(PureWindowsPath(build_name), obj_files)
    obj_files = convert_posix_to_windows(obj_files)

    file_pairs = zip(to_str(src_files), to_str(obj_files))
    for src_file, obj_file in file_pairs:
        writer.build(
            outputs=obj_file,
            rule="compile",
            inputs=src_file,
            variables={
                "compiler": compiler,
                "includes": includes,
                "flags": cxx_flags,
                "defines": defines,
            },
        )
        writer.newline()

    return obj_files


def generate_linker_args(build: Dict, parsed_toml: Dict):
    lib_infos = commonbuilds.get_required_library_information(build, parsed_toml)

    (
        _,
        link_libraries,
        __,
        ___,
    ) = get_library_information(lib_infos)

    # Requires Libraries:
    #
    requires_libraries = PrefixLibrary(link_libraries)
    items = [info.path for info in lib_infos]
    requires_library_paths = PrefixLibraryPath(wrap_quotes(items))

    # External Libraries:
    #
    (
        external_libraries_names,
        external_libraries_paths,
    ) = get_external_libraries_information(build)

    external_libraries_names = PrefixLibrary(PostFixLib(external_libraries_names))
    external_libraries_paths = PrefixLibraryPath(wrap_quotes(external_libraries_paths))

    # Reference Libraries:
    #
    (
        ref_libraries,
        ref_library_paths,
    ) = commonbuilds.get_reference_library_information(build, parsed_toml)

    ref_libraries = PrefixLibrary(PostFixLib(ref_libraries))
    ref_library_paths = PrefixLibraryPath(wrap_quotes(convert_posix_to_windows(ref_library_paths)))

    linker_args = (
            requires_library_paths
            + external_libraries_paths
            + ref_library_paths
            + requires_libraries
            + external_libraries_names
            + ref_libraries
    )
    return linker_args


def run_build(build: Dict, parsed_toml: Dict, ninja_writer: Writer, args):
    # TODO forward args
    the_build = BuildTypes[build["buildRule"]]
    build_name = build["name"]
    build_dir = build["build_dir"]

    build_path = build_dir / build_name
    build_path.mkdir(parents=True, exist_ok=True)

    build["buildPath"] = build_path

    if the_build == BuildTypes.staticLibrary:
        build_static_library(
            ninja_writer,
            build,
            parsed_toml,
            windows_add_static_library_naming_convention,
        )
    elif the_build == BuildTypes.executable:
        build_executable(ninja_writer, build, parsed_toml)
    elif the_build == BuildTypes.dynamicLibrary:
        build_dynamic_library(ninja_writer, build, parsed_toml)
    elif the_build == BuildTypes.headerOnly:
        pass
    elif the_build == BuildTypes.libraryReference:
        pass
    else:
        raise RuntimeError(f"Unknown build type {the_build}.")


def build_static_library(pfw: Writer,
                         build: Dict,
                         parsed_toml: Dict,
                         lib_name_func: Callable[[str], str]):
    build_name = build["name"]

    includes = get_includes_for_build(build, parsed_toml)
    obj_files = add_compile_rule(pfw, build, parsed_toml, includes)

    library_name = lib_name_func(build["outputName"])
    relative_output_name = str(PureWindowsPath(build_name) / library_name)

    _, archiver, cxx_flags, defines, __, ___ = commonbuilds.get_toolchain_and_flags(
        build, parsed_toml
    )
    defines = PrefixHashDefine(defines)

    pfw.build(
        outputs=relative_output_name,
        rule="archive",
        inputs=to_str(obj_files),
        variables={
            "archiver": archiver,
            "includes": includes,
            "flags": cxx_flags,
            "defines": defines,
        },
    )

    pfw.newline()
    pfw.build(rule="phony", inputs=relative_output_name, outputs=library_name)
    pfw.build(rule="phony", inputs=library_name, outputs=build_name)
    pfw.newline()


def build_executable(pfw: Writer, build: Dict, parsed_toml: Dict):
    build_name = build["name"]

    _, __, cxx_flags, defines, linker, linker_args = commonbuilds.get_toolchain_and_flags(
        build, parsed_toml
    )
    defines = PrefixHashDefine(defines)

    includes = get_includes_for_build(build, parsed_toml)
    obj_files = add_compile_rule(pfw, build, parsed_toml, includes)

    lib_infos = commonbuilds.get_required_library_information(build, parsed_toml)

    (_, __, implicit_exps, implicit_libs,) = get_library_information(lib_infos)

    linker_args = generate_linker_args(build, parsed_toml)

    exe_name = build["outputName"]
    if USING_RELATIVE_OUTPUTS:
        build_output = str(PureWindowsPath(build_name) / exe_name)
    else:
        build_output = exe_name

    pfw.build(
        rule="exe",
        inputs=to_str(obj_files),
        implicit=implicit_exps + implicit_libs,
        outputs=build_output,
        variables={
            "compiler": linker,
            "includes": includes,
            "flags": cxx_flags,
            "defines": " ".join(defines),
            "exe_name": build_output,
            "linker_args": " ".join(linker_args),
        },
    )
    pfw.newline()
    if USING_RELATIVE_OUTPUTS:
        pfw.build(rule="phony", inputs=build_output, outputs=exe_name)
    pfw.build(rule="phony", inputs=exe_name, outputs=build_name)
    pfw.newline()


def build_dynamic_library(pfw: Writer, build: Dict, parsed_toml: Dict):
    build_name = build["name"]

    extra_flags = ["/DEXPORT_DLL_PUBLIC"]

    compiler, _, cxxflags, defines, __, ___ = commonbuilds.get_toolchain_and_flags(
        build, parsed_toml
    )
    defines = PrefixHashDefine(defines)

    includes = get_includes_for_build(build, parsed_toml)
    obj_files = add_compile_rule(pfw, build, parsed_toml, includes, extra_flags)

    lib_infos = commonbuilds.get_required_library_information(build, parsed_toml)

    (_, __, implicit_exps, implicit_libs,) = get_library_information(lib_infos)

    linker_args = generate_linker_args(build, parsed_toml)

    lib, exp = convert_to_implicit_library_files(build["outputName"])

    if USING_RELATIVE_OUTPUTS:
        implicit_outputs = [
            str(PureWindowsPath(build_name) / lib),
            str(PureWindowsPath(build_name) / exp),
        ]
    else:
        implicit_outputs = [lib, exp]

    library_name = windows_add_dynamic_library_naming_convention(
        build["outputName"]
    )

    if USING_RELATIVE_OUTPUTS:
        build_output = str(PureWindowsPath(build_name) / library_name)
    else:
        build_output = library_name

    pfw.build(
        rule="shared",
        inputs=to_str(obj_files),
        implicit=implicit_exps + implicit_libs,
        outputs=build_output,
        implicit_outputs=implicit_outputs,
        variables={
            "compiler": compiler,
            "includes": includes,
            "flags": " ".join(cxxflags),
            "defines": " ".join(defines),
            "lib_name": build_output,
            "linker_args": " ".join(linker_args),
        },
    )
    pfw.newline()
    if USING_RELATIVE_OUTPUTS:
        pfw.build(rule="phony", inputs=build_output, outputs=library_name)
    pfw.build(rule="phony", inputs=library_name, outputs=build_name)
    pfw.newline()


def log_build_information(build):
    build_name = build["name"]
    cxxflags = build["flags"]
    defines = build["defines"]
    includes = build["includes"]
    library_paths = build["libraryPaths"]
    output = build["outputName"]

    print(f"Running build: f{build_name}")
    print(f"CXXFLAGS: {cxxflags}")
    print(f"DEFINES: {defines}")
    print(f"INCLUDE_PATHS: {includes}")
    print(f"LIBRARY_PATHS: {library_paths}")
    print(f"OUTPUT NAME: {output}")
    print("")



import pprint
import sys
from pathlib import Path
from typing import Union, List

import cerberus

from aim_build.commonbuilds import BuildTypes
from aim_build.utils import to_native_path


class UniqueNameChecker:
    def __init__(self):
        self.name_lookup = []

    def check(self, field, value, error):
        if value in self.name_lookup:
            error(
                field,
                f'The build name must be unique. "{value}" has already been used.',
            )
        else:
            self.name_lookup.append(value)


# TODO: Can DefinesPrefixChecker be a function?
class DefinesPrefixChecker:
    def check(self, field, defines: List[str], error):
        for value in defines:
            if value.startswith("-D"):
                error(
                    field,
                    f"Unnecessary -D prefix in {field}: {defines}. Aim will add this automatically.",
                )


class RequiresExistChecker:
    def __init__(self, document):
        self.doc = document

    def check(self, field, requires, error):
        builds = self.doc["builds"]
        for value in requires:
            for build in builds:
                if value == build["name"]:
                    break
            else:
                error(field, f"Build name does not exist: {value}")


class DirectoryPathChecker:
    def __init__(self, project_dir):
        self.project_dir = project_dir

    def check(self, field, paths, error):
        abs_paths = [Path(path) for path in paths if Path(path).is_absolute() is True]
        rel_paths = [Path(path) for path in paths if Path(path).is_absolute() is False]
        rel_paths = [(self.project_dir / path) for path in rel_paths]

        all_paths = abs_paths + rel_paths

        for path in all_paths:
            if not path.is_dir():
                error(field, f'Path is not a directory: "{str(path)}"')
                break

            if not path.exists():
                error(field, f'Path does not exist: "{str(path)}"')
                break


class SrcPathsChecker:
    def __init__(self, project_dir):
        self.project_dir = project_dir

    def check(self, field, paths, error):
        paths = [to_native_path(path) for path in paths]

        # Resolve relative paths to the build directory, and leave absolute paths alone.
        #
        abs_paths = [path for path in paths if path.is_absolute() is True]
        rel_paths = [(self.project_dir / path) for path in paths if path.is_absolute() is False]
        all_paths = abs_paths + rel_paths

        for path in all_paths:
            # Historically, paths to directories were allowed, but I've decided that it is better to specify directories
            # using a glob with a file extension, as this means we can support any file type without having to
            # explicitly manage the file types that we support.

            # Remember paths can now be specific paths to files, or globs of the form <path>/*.<extension>.
            # We don't allow recursive globs. <path>/**/*.<extension>
            if path.stem == "**":
                error(field, f'Recursive globs are not supported: "{str(path)}"')
                return

            elif path.stem == "*":
                parent = path.parent
                if not parent.exists():
                    error(field, f'The parent glob directory does not exist: "{str(parent)}"')
                    break

                if len(list(parent.glob(path.name))) == 0:
                    error(field, f'The glob does not match any files: "{str(path)}"')
                    break

            elif path.is_dir():
                error(field, f'Src path is a directory. Src paths should be a glob or a specific file.: "{str(path)}"')
                break

            elif not path.exists():
                error(field, f'Path does not exist: "{str(path)}"')
                break


class AimCustomValidator(cerberus.Validator):
    def __init__(self, *args, **kwargs):
        super(AimCustomValidator, self).__init__(*args, **kwargs)
        self.name_lookup = []

    def check_output_is_unique(self, field, value):
        if value in self.name_lookup:
            self._error(
                field,
                f'The output name must be unique. "{value}" has already been used.',
            )
        else:
            self.name_lookup.append(value)

    def _check_with_output_naming_convention(self, field, value: Union[str, list]):
        # Only check the convention when building libraries. Exes need more flexibility when building for Arduino etc.

        library_types = [BuildTypes.staticLibrary.name, BuildTypes.dynamicLibrary.name]
        if self.document["buildRule"] not in library_types:
            return

        def check_convention(_field, _value):
            the_errors = []
            if _value.startswith("lib"):
                the_error_str = f"Unnecessary 'lib' prefix in {_value}. Aim will add this automatically."
                the_errors.append(the_error_str)

            suffix = Path(_value).suffix
            if suffix:
                the_error_str = (
                    f'Unecessary suffix "{suffix}". Aim will add this automatically.'
                )
                the_errors.append(the_error_str)

            return the_errors

        # Bit of a hack so strings go through the same code path as lists.
        if isinstance(value, str):
            value = [value]

        for item in value:
            # self.check_output_is_unique(field, item)
            errors = check_convention(field, item)

            if errors:
                plural = ""
                if len(errors) > 1:
                    plural = "s"

                error_str = f"Naming convention error{plural}: {item}. " + " ".join(
                    errors
                )
                self._error(field, error_str)


def target_schema(document, project_dir):
    unique_name_checker = UniqueNameChecker()
    requires_exist_checker = RequiresExistChecker(document)
    source_path_checker = SrcPathsChecker(project_dir)
    directory_path_checker = DirectoryPathChecker(project_dir)
    defines_checker = DefinesPrefixChecker()

    schema = {
        "compiler": {"required": True, "type": "string"},
        "archiver": {"required": True, "type": "string"},
        "compilerFrontend": {
            "required": True,
            "type": "string",
            "allowed": ["msvc", "gcc", "osx"],
        },
        "flags": {
            "type": "list",
            "schema": {"type": "string"},
            "empty": False
        },
        "defines": {
            "type": "list",
            "schema": {"type": "string"},
            "empty": False,
            "check_with": defines_checker.check,
        },
        "projectRoot": {
            "required": True,
            "type": "string",
            "empty": False
        },
        "builds": {
            "required": True,
            "type": "list",
            "schema": {
                "type": "dict",
                "schema": {
                    "name": {
                        "required": True,
                        "type": "string",
                        "check_with": unique_name_checker.check,
                    },
                    "buildRule": {
                        "required": True,
                        "type": "string",

                        # Validates if exactly one of the provided constraints applies.
                        "oneof": [
                            {
                                # No output, so exclude output name.
                                # Not a dynamic library so exclude visibility and dynamicLoading
                                "excludes": ["outputName", "visibility", "dynamicLoading"],
                                "allowed": ["headerOnly", "libraryReference"],
                            },
                            {
                                # Generates an output so make outputName required.
                                # Not a dynamic library so exclude visibility and dynamicLoading.
                                "excludes": ["visibility", "dynamicLoading"],
                                "dependencies": ["outputName"],
                                "allowed": ["executable", "staticLibrary"],
                            },
                            {
                                # Generates an output so make outputName required.
                                # Allows visibility if frontend is gcc or osx.
                                "dependencies": ["outputName"],
                                "allowed": ["dynamicLibrary"],
                            },
                        ],
                    },
                    "visibility": {
                        "required": False,
                        "type": "string",
                        "allowed": [
                            "hidden",
                            "default"
                        ],
                        "dependencies": {"^compilerFrontend": ["gcc", "osx"]},
                    },
                    "dynamicLoading": {
                        "required": False,
                        "type": "boolean",
                    },
                    "compiler": {
                        "required": False,
                        "type": "string",
                        "dependencies": {
                            "buildRule": ["executable", "staticLibrary", "dynamicLibrary"]
                        },
                    },
                    "linker": {
                        "required": False,
                        "type": "string",
                        "dependencies": {
                            "buildRule": ["executable"]
                        },
                    },
                    "linker_flags": {
                        "required": False,
                        "type": "list",
                        "dependencies": {
                            "buildRule": ["executable"]
                        },
                    },
                    "defines": {
                        "type": "list",
                        "schema": {"type": "string"},
                        "empty": False,
                        "check_with": defines_checker.check,
                        "dependencies": {
                            "buildRule": ["executable", "staticLibrary", "dynamicLibrary"]
                        },
                    },
                    "flags": {
                        "type": "list",
                        "schema": {"type": "string"},
                        "empty": False,
                        "dependencies": {
                            "buildRule": ["executable", "staticLibrary", "dynamicLibrary"]
                        },
                    },
                    "requires": {
                        "type": "list",
                        "empty": False,
                        "schema": {"type": "string"},
                        "check_with": requires_exist_checker.check,
                        "dependencies": {
                            "buildRule": ["executable", "staticLibrary", "dynamicLibrary"]
                        },
                    },
                    # Required but the requirement is handled by build rule.
                    "outputName": {
                        "type": "string",
                        "empty": False,
                        "check_with": "output_naming_convention",
                    },
                    "sourceFiles": {
                        "required": False,
                        "empty": False,
                        "type": "list",
                        "schema": {"type": "string"},
                        "check_with": source_path_checker.check,
                        "dependencies": {
                            "buildRule": ["executable", "staticLibrary", "dynamicLibrary"]
                        },
                    },
                    "includePaths": {
                        "type": "list",
                        "empty": False,
                        "schema": {"type": "string"},
                        "check_with": directory_path_checker.check,
                    },
                    "systemIncludePaths": {
                        "type": "list",
                        "empty": False,
                        "schema": {"type": "string"},
                        "check_with": directory_path_checker.check,
                        "dependencies": {"compilerFrontend": "gcc"},
                    },
                    "localIncludePaths": {
                        "type": "list",
                        "empty": False,
                        "schema": {"type": "string"},
                        "check_with": directory_path_checker.check,
                        "dependencies": {"compilerFrontend": "gcc"},
                    },
                    "libraryPaths": {
                        "type": "list",
                        "empty": False,
                        "schema": {"type": "string"},
                        # you can't check the library dirs as they may not exist if the project not built before.
                        # "check_with": path_checker.check,
                        "dependencies": {
                            "buildRule": ["executable", "dynamicLibrary", "libraryReference"]
                        },
                    },
                    "libraries": {
                        "type": "list",
                        "empty": False,
                        "schema": {"type": "string"},
                        "check_with": "output_naming_convention",
                        "dependencies": {
                            "buildRule": ["executable", "dynamicLibrary", "libraryReference"]
                        },
                    },
                },
            },
        },
    }

    validator = AimCustomValidator()
    validator.validate(document, schema)

    pretty = pprint.PrettyPrinter(indent=2, width=100)

    # TODO: Handle schema errors. https://docs.python-cerberus.org/en/stable/errors.html
    if validator.errors:
        for key, value in validator.errors.items():
            if key != "builds":
                print(f'Error for field "{key}"')
                pretty.pprint(f"{value}")
                print()

        builds_errors = validator.errors.get("builds", {})
        if builds_errors:
            assert len(builds_errors) == 1, "Build error list size is greater than 1."
            for key, value in builds_errors[0].items():
                builds = document["builds"]
                the_build = builds[key]
                the_build_name = the_build["name"]

                print(f'Error in build: "{the_build_name}"')
                assert (
                        len(value) == 1
                ), "Length is not 1. Not sure if it can ever be more than."
                pretty.pprint(value[0])
                print()
        sys.exit(-1)



import functools
from pathlib import PurePosixPath, Path
from typing import Dict, Tuple, Callable, List

from ninja_syntax import Writer

from aim_build import commonbuilds
from aim_build.commonbuilds import BuildTypes
from aim_build.typedefs import StringList, PurePathList
from aim_build.utils import (
    prefix,
    src_to_o,
    to_path_root,
    prepend_paths,
    to_str,
    escape_path,
    wrap_quotes
)

PrefixIncludePath = functools.partial(prefix, "-I")
PrefixSystemIncludePath = functools.partial(prefix, "-isystem")
PrefixQuoteIncludePath = functools.partial(prefix, "-iquote")
PrefixLibraryPath = functools.partial(prefix, "-L")
PrefixLibrary = functools.partial(prefix, "-l")
PrefixHashDefine = functools.partial(prefix, "-D")
ToObjectFiles = src_to_o


def add_compile(nfw: Writer):
    command = "$compiler $defines $flags -MMD -MF deps.d $includes -c $in -o $out"
    nfw.rule(
        name="compile",
        description="Compiles source files into object files",
        deps="gcc",
        depfile="deps.d",
        command=command,
    )
    nfw.newline()


def add_ar(nfw: Writer):
    nfw.rule(
        name="archive",
        description="Combine object files into an archive",
        command="$archiver crs $out $in",
    )
    nfw.newline()


def add_exe(nfw: Writer):
    # TODO: origin should only really be added when we need to link against an so.
    command = "$compiler $defines $flags $includes $in -o $out $linker_args"
    nfw.rule(name="exe", description="Builds an executable.", command=command)
    nfw.newline()


def add_shared(nfw: Writer):
    command = "$compiler $defines -shared $visibility -fPIC $flags $includes $in -o $out $linker_args"
    nfw.rule(name="shared", description="Builds a shared library.", command=command)
    nfw.newline()


# TODO: Should take version strings as well?
def linux_add_static_library_naming_convention(library_name: str) -> str:
    return f"lib{library_name}.a"


def linux_add_dynamic_library_naming_convention(library_name: str) -> str:
    return f"lib{library_name}.so"


# def linux_add_exe_naming_convention(exe_name: str) -> str:
#     return f"{exe_name}.exe"


def get_quote_include_paths(build: Dict, build_dir: PurePosixPath) -> PurePathList:
    rel_path = to_path_root(build_dir)

    include_paths = build.get("localIncludePaths", [])
    includes = [rel_path / PurePosixPath(path) for path in include_paths]

    return includes


def get_system_include_paths(build: Dict) -> PurePathList:
    paths = build.get("systemIncludePaths", [])
    paths = [PurePosixPath(path) for path in paths]
    return paths


def convert_strings_to_paths(paths):
    return [PurePosixPath(path) for path in paths]


def get_includes_for_build(build: Dict, parsed_toml: Dict) -> StringList:
    requires = [build["name"]] + build.get("requires", [])

    include_paths = set()
    system_include_paths = set()
    quote_include_paths = set()

    project_root = PurePosixPath(parsed_toml["projectRoot"])

    for required in requires:
        the_dep = commonbuilds.find_build(required, parsed_toml["builds"])

        includes = the_dep.get("includePaths", [])
        includes = convert_strings_to_paths(includes)
        includes = commonbuilds.get_include_paths(includes, project_root)
        include_paths.update(includes)

        quote_includes = get_quote_include_paths(the_dep, build["build_dir"])
        quote_include_paths.update(quote_includes)

        system_includes = get_system_include_paths(the_dep)
        system_include_paths.update(system_includes)

    include_paths = wrap_quotes(list(include_paths))
    system_include_paths = wrap_quotes(list(system_include_paths))
    quote_include_paths = wrap_quotes(list(quote_include_paths))

    include_paths.sort()
    system_include_paths.sort()
    quote_include_paths.sort()

    include_args = PrefixIncludePath(include_paths)
    system_include_args = PrefixSystemIncludePath(system_include_paths)
    quote_args = PrefixQuoteIncludePath(quote_include_paths)

    return include_args + system_include_args + quote_args


def get_external_libraries_paths(build: Dict) -> StringList:
    directory = build["directory"]
    library_paths = build.get("libraryPaths", [])
    result = []
    for path in library_paths:
        path = PurePosixPath(path)
        if not path.is_absolute():
            path = directory / path

        result.append(str(path))
    result = wrap_quotes(result)
    return result

    # library_paths = wrap_quotes(library_paths)
    # library_paths = prepend_paths(directory, library_paths)
    # return library_paths


def get_external_libraries_names(build: Dict) -> Tuple[StringList, StringList]:
    libraries = build.get("libraries", [])
    link_libraries = PrefixLibrary(libraries)
    return libraries, link_libraries


def get_external_libraries_information(build: Dict) -> Tuple[StringList, StringList]:
    libraries, _ = get_external_libraries_names(build)
    library_paths = get_external_libraries_paths(build)
    return libraries, library_paths


def get_src_for_build(build: Dict, parsed_toml: Dict) -> List[PurePosixPath]:
    files = commonbuilds.get_src_files(build, parsed_toml)
    return convert_strings_to_paths(files)


def add_compile_rule(
        writer: Writer,
        build: Dict,
        target_file: Dict,
        includes: StringList,
        extra_flags: StringList = None,
):
    build_name = build["name"]

    compiler, _, cxx_flags, defines, __, ___ = commonbuilds.get_toolchain_and_flags(
        build, target_file
    )
    defines = PrefixHashDefine(defines)

    if extra_flags:
        cxx_flags = extra_flags + cxx_flags

    src_files = get_src_for_build(build, target_file)
    obj_files = ToObjectFiles(src_files)
    obj_files = prepend_paths(Path(build_name), obj_files)

    file_pairs = zip(to_str(src_files), to_str(obj_files))
    for src_file, obj_file in file_pairs:
        writer.build(
            outputs=obj_file,
            rule="compile",
            inputs=src_file,
            variables={
                "compiler": compiler,
                "includes": includes,
                "flags": cxx_flags,
                "defines": defines,
            },
        )
        writer.newline()

    return obj_files


def get_rpath(build: Dict, parsed_toml: Dict) -> str:
    # Good blog post about rpath:
    # https://medium.com/@nehckl0/creating-relocatable-linux-executables-by-setting-rpath-with-origin-45de573a2e98
    requires = build.get("requires", [])
    library_names = set()

    # find_build_types("dynamicLib", parsed_toml["builds"])
    # TODO: replace the below with the above.
    for required in requires:
        the_dep = commonbuilds.find_build(required, parsed_toml["builds"])
        build_type = BuildTypes[the_dep["buildRule"]]
        if build_type == BuildTypes.dynamicLibrary:
            library_names.update([the_dep["name"]])

    # Note, thinking aloud here as not 100% sure this is the correct thing to do.
    # The current thought process is that the origin allows something like an executable to find the libraries it needs.
    # Therefore, the path to these dependencies is simply just up one directory, which takes us to the build directory,
    # and then down into the build dependency folder.

    library_names = list(library_names)
    library_names.sort()
    relative_paths = [PurePosixPath("..") / lib_path for lib_path in library_names]

    relative_paths = [f"$$ORIGIN/{rel_path}" for rel_path in relative_paths]
    relative_paths = ["$$ORIGIN"] + relative_paths

    relative_paths_string = escape_path(":".join(relative_paths))
    return f"-Wl,-rpath='{relative_paths_string}'"


def generate_linker_args(build, parsed_toml):
    lib_infos = commonbuilds.get_required_library_information(build, parsed_toml)

    rpath = get_rpath(build, parsed_toml)

    # Requires Libraries:
    #
    requires_libraries = PrefixLibrary([info.name for info in lib_infos])
    requires_library_paths = [info.path for info in lib_infos]
    requires_library_paths = wrap_quotes(requires_library_paths)
    requires_library_paths = PrefixLibraryPath(requires_library_paths)

    # External Libraries:
    #
    (
        external_libraries_names,
        external_libraries_paths,
    ) = get_external_libraries_information(build)

    external_libraries_names = PrefixLibrary(external_libraries_names)
    external_libraries_paths = wrap_quotes(external_libraries_paths)
    external_libraries_paths = PrefixLibraryPath(external_libraries_paths)

    # Reference Libraries:
    #
    (
        ref_libraries,
        ref_library_paths,
    ) = commonbuilds.get_reference_library_information(build, parsed_toml)

    ref_libraries = PrefixLibrary(ref_libraries)
    ref_library_paths = wrap_quotes(ref_library_paths)
    ref_library_paths = PrefixLibraryPath(ref_library_paths)

    linker_args = (
            [rpath]
            + requires_library_paths
            + external_libraries_paths
            + ref_library_paths
            + requires_libraries
            + external_libraries_names
            + ref_libraries
    )
    return linker_args


def run_build(build: Dict, parsed_toml: Dict, ninja_writer: Writer, args):
    # TODO forward args
    the_build = BuildTypes[build["buildRule"]]
    build_name = build["name"]
    build_dir = build["build_dir"]

    build_path = build_dir / build_name
    build_path.mkdir(parents=True, exist_ok=True)

    build["buildPath"] = build_path

    if the_build == BuildTypes.staticLibrary:
        build_static_library(
            ninja_writer,
            build,
            parsed_toml,
            linux_add_static_library_naming_convention,
        )
    elif the_build == BuildTypes.executable:
        build_executable(ninja_writer, build, parsed_toml)
    elif the_build == BuildTypes.dynamicLibrary:
        build_dynamic_library(ninja_writer, build, parsed_toml)
    elif the_build == BuildTypes.headerOnly:
        pass
    elif the_build == BuildTypes.libraryReference:
        pass
    else:
        raise RuntimeError(f"Unknown build type {the_build}.")


def build_static_library(pfw: Writer, build: Dict, parsed_toml: Dict, lib_name_func: Callable[[str], str]):
    build_name = build["name"]

    includes = get_includes_for_build(build, parsed_toml)
    obj_files = add_compile_rule(pfw, build, parsed_toml, includes)

    library_name = lib_name_func(build["outputName"])
    relative_output_name = str(PurePosixPath(build_name) / library_name)

    _, archiver, cxx_flags, defines, __, ___ = commonbuilds.get_toolchain_and_flags(
        build, parsed_toml
    )
    defines = PrefixHashDefine(defines)

    pfw.build(
        outputs=relative_output_name,
        rule="archive",
        inputs=to_str(obj_files),
        variables={
            "archiver": archiver,
            "includes": includes,
            "flags": cxx_flags,
            "defines": defines,
        },
    )

    pfw.newline()
    pfw.build(rule="phony", inputs=relative_output_name, outputs=library_name)
    pfw.build(rule="phony", inputs=library_name, outputs=build_name)
    pfw.newline()


def build_executable(pfw: Writer, build: Dict, parsed_toml: Dict):
    build_name = build["name"]

    _, __, ___, defines, linker, linker_flags = commonbuilds.get_toolchain_and_flags(
        build, parsed_toml
    )
    defines = PrefixHashDefine(defines)

    includes = get_includes_for_build(build, parsed_toml)
    obj_files = add_compile_rule(pfw, build, parsed_toml, includes)

    lib_infos = commonbuilds.get_required_library_information(build, parsed_toml)
    linker_args = generate_linker_args(build, parsed_toml)

    full_library_names = commonbuilds.get_full_library_name_convention(
        lib_infos,
        linux_add_static_library_naming_convention,
        linux_add_dynamic_library_naming_convention,
    )

    exe_name = build["outputName"]
    relative_output_name = str(Path(build_name) / exe_name)
    pfw.build(
        rule="exe",
        outputs=relative_output_name,
        inputs=to_str(obj_files),
        implicit=full_library_names,
        variables={
            "compiler": linker,
            "includes": includes,
            "flags": linker_flags,
            "defines": defines,
            "linker_args": " ".join(linker_args),
        },
    )
    pfw.newline()
    pfw.build(rule="phony", inputs=relative_output_name, outputs=exe_name)
    pfw.build(rule="phony", inputs=exe_name, outputs=build_name)
    pfw.newline()


def build_dynamic_library(pfw: Writer, build: Dict, parsed_toml: Dict):
    build_name = build["name"]

    vis = build.get("visibility", None)
    if vis is None:
        vis = "hidden"
    vis_flags = f"-fvisibility={vis}"

    extra_flags = ["-DEXPORT_DLL_PUBLIC", "-fPIC", vis_flags]

    compiler, _, cxxflags, defines, __, ___ = commonbuilds.get_toolchain_and_flags(
        build, parsed_toml
    )
    defines = PrefixHashDefine(defines)

    includes = get_includes_for_build(build, parsed_toml)
    obj_files = add_compile_rule(pfw, build, parsed_toml, includes, extra_flags)

    lib_infos = commonbuilds.get_required_library_information(build, parsed_toml)
    linker_args = generate_linker_args(build, parsed_toml)

    full_library_names = commonbuilds.get_full_library_name_convention(
        lib_infos,
        linux_add_static_library_naming_convention,
        linux_add_dynamic_library_naming_convention,
    )

    library_name = linux_add_dynamic_library_naming_convention(build["outputName"])
    relative_output_name = str(PurePosixPath(build_name) / library_name)

    pfw.build(
        rule="shared",
        inputs=to_str(obj_files),
        implicit=full_library_names,
        outputs=relative_output_name,
        variables={
            "compiler": compiler,
            "includes": includes,
            "flags": " ".join(cxxflags),
            "defines": " ".join(defines),
            "lib_name": library_name,
            "linker_args": " ".join(linker_args),
        },
    )
    pfw.newline()
    pfw.build(rule="phony", inputs=relative_output_name, outputs=library_name)
    pfw.build(rule="phony", inputs=library_name, outputs=build_name)
    pfw.newline()


def log_build_information(build):
    build_name = build["name"]
    cxx_flags = build["global_flags"] + build.get("flags", [])
    defines = build["global_defines"] + build.get("defines", [])
    includes = build["includes"]
    library_paths = build["libraryPaths"]
    output = build["outputName"]

    print(f"Running build: f{build_name}")
    print(f"FLAGS: {cxx_flags}")
    print(f"DEFINES: {defines}")
    print(f"INCLUDE_PATHS: {includes}")
    print(f"LIBRARY_PATHS: {library_paths}")
    print(f"OUTPUT NAME: {output}")
    print("")



from pathlib import Path, PurePath
from typing import List, TypeVar

# Convenient typedefs
#
StringList = List[str]
PathList = List[Path]
PurePathList = List[PurePath]

# A type variable so function types can vary.
#
T = TypeVar("T")



import itertools
import os
from pathlib import Path, PurePath, PurePosixPath
from typing import List, Union, Iterable

from aim_build.typedefs import PathList, PurePathList, StringList, T


def src_to_obj(files) -> StringList:
    return [x.stem + ".obj" for x in files]


def src_to_o(files) -> StringList:
    return [x.stem + ".o" for x in files]


def to_str(paths) -> StringList:
    return [str(x) for x in paths]


def to_paths(string_paths) -> PathList:
    return [Path(x) for x in string_paths]


def glob(glob_string, paths: PathList) -> List[PathList]:
    return [list(x.glob(glob_string)) for x in paths]


def flatten(list_of_lists: List[List[T]]) -> List[T]:
    return list(itertools.chain.from_iterable(list_of_lists))


def prefix(the_prefix, paths) -> StringList:
    return [the_prefix + str(x) for x in paths]


def postfix(the_postfix, paths) -> StringList:
    return [str(x) + the_postfix for x in paths]


def wrap_quotes(paths: Union[PurePathList, StringList, str]) -> StringList:
    if isinstance(paths, str):
        str_path = str(paths)
        return [f'"{str_path}"']

    str_paths = to_str(paths)
    return [f'"{x}"' for x in str_paths]


def suffix(the_suffix, paths) -> StringList:
    return [str(x) + the_suffix for x in paths]


def prepend_paths(base_path: PurePath, other_paths: Union[List[PurePath], StringList]):
    # Don't need to check if `the_path` is absolute. If it is, the the result of `base_path / the_path` is just the
    # `the_path`. So it does the right thing, even though you might not expect it.
    return [base_path / the_path for the_path in other_paths]


def resolve(paths: PathList):
    return [path.resolve() for path in paths]


def escape_path(word):
    return word.replace("$ ", "$$ ").replace(" ", "$ ").replace(":", "$:")


def relpath(src_path: PurePosixPath, dst_path: PurePosixPath):
    result = os.path.relpath(str(src_path), str(dst_path))
    result = result.replace("\\", "/")
    return PurePosixPath(result)


def relpaths(src_paths: Union[PurePosixPath, List[PurePosixPath]], dst_path: PurePosixPath) -> List[PurePosixPath]:
    if isinstance(src_paths, Iterable):
        return [relpath(src_path, dst_path) for src_path in src_paths]

    return [relpath(src_paths, dst_path)]


def to_path_root(build_path: PurePath) -> PurePath:
    rel_path = PurePosixPath()
    for _ in range(len(build_path.parents)):
        rel_path = rel_path / ".."
    return rel_path


def to_native_path(path: PurePath) -> Path:
    # Have to be careful when converting from posix to windows paths.
    # If you have a posix path object pointing to C:/some/path/somewhere and simply do:
    #   Path(posix_path)
    # Then the result will be C:some\path\somewhere - the \ after the C: drive is missing.
    # However the conversion works, if you convert to a string before constructing the native path object.
    return Path(str(path))


def to_pure_posix_path(path: PurePath) -> PurePosixPath:
    # This function handles some common errors when converting from Windows paths back to PurePosixPaths.
    return PurePosixPath(str(path).replace("\\", "/"))



import argparse
import os
import shutil
import subprocess
import sys
import zipfile
from pathlib import Path

from ninja_syntax import Writer
from tabulate import tabulate

from aim_build import gccbuilds
from aim_build import msvcbuilds
from aim_build.common import DEMO_ZIP_FILE_NAME
from aim_build.commonbuilds import find_build, BuildTypes
from aim_build.schema import target_schema
from aim_build.version import __version__


def run_ninja(working_dir, build_name):
    command = ["ninja", "-C", str(working_dir), "-v", build_name]
    # command_str = " ".join(command)
    # print(f'Executing "{command_str}"')

    with subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE) as process:
        for line in iter(process.stdout.readline, b""):
            sys.stdout.write(line.decode("utf-8"))
        for line in iter(process.stderr.readline, b""):
            sys.stderr.write(line.decode("utf-8"))

        ret_code = process.returncode

    return ret_code


def entry():
    # print("DEV")
    script_path = Path(__file__).parent

    # TODO: Get version automatically from the pyproject.toml file.
    parser = argparse.ArgumentParser(prog="aim", description=f"Version {__version__}")

    sub_parser = parser.add_subparsers(dest="command", help="Commands")

    # Version command
    # -------------
    # sub_parser.add_parser(name="version", help="Displays the version number")

    # Init command
    # -------------
    init_parser = sub_parser.add_parser(name="init", help="creates a project structure")

    init_parser.add_argument("--demo-files",
                             help="create additional demo files",
                             action="store_true")

    # Target command
    # --------------
    target_parser = sub_parser.add_parser(name="target",
                                          help="[-h] initiate target commands {build, list, clobber}")

    target_parser.add_argument("path",
                               type=str,
                               help="path to directory containing target.py")

    target_sub = target_parser.add_subparsers(dest="target")

    build_parser = target_sub.add_parser(name="build",
                                         help="[-h] executes a build")

    build_parser.add_argument("build",
                              type=str,
                              help="the build name")

    build_parser.add_argument(
        "--skip-ninja-regen",
        help="by-pass the ninja file generation step",
        action="store_true",
    )

    # TODO: Get this working again.
    # build_parser.add_argument(
    #     "--profile-build",
    #     help="forwards -ftime-trace to the compiler for emitting build profile information."
    #          " View using chome://tracing.",
    #     action="store_true",
    # )

    # TODO: Get this working again.
    # build_parser.add_argument(
    #     "--args", help="additional arguments forwarded to the compiler", nargs="*"
    # )

    run_parser = target_sub.add_parser(name="run",
                                       help="[-h] runs the build if it is an executable")

    run_parser.add_argument("build",
                            type=str,
                            help="the build name")

    run_parser.add_argument("args",
                            nargs=argparse.REMAINDER,
                            help="arguments forwarded to the executable")

    target_sub.add_parser(name="list",
                          help="display the builds")

    target_sub.add_parser(name="clobber",
                          help="deletes all build artifacts")

    args = parser.parse_args()
    mode = args.command
    if mode == "init":
        if args.demo_files:
            print("Initialising from demo project...")
            relative_dir = "demo/Calculator"
        else:
            relative_dir = "demo/Empty"

        zip_path = script_path / f"{DEMO_ZIP_FILE_NAME}"

        assert zip_path.exists(), f"Failed to find demo zip files: {str(zip_path)}"
        with zipfile.ZipFile(str(zip_path)) as zip_file:
            run_init(zip_file, relative_dir)
    elif mode == "target":
        target = args.target
        path = args.path
        if target == "build":
            args.args = None
            args.profile_build = None
            forwarding_args = [] if args.args is None else args.args
            if args.profile_build and "-ftime-trace" not in forwarding_args:
                forwarding_args.append("-ftime-trace")

            ret_code = run_build(args.build, path, args.skip_ninja_regen, forwarding_args)
            sys.exit(ret_code)
        elif target == "list":
            run_list(path)
        elif target == "clobber":
            run_clobber(path)
        elif target == "run":
            run_run(path, args)

    else:
        parser.print_help(sys.stdout)


def run_run(path, args):
    target_file = load_target_file(path)

    the_build = find_build(args.build, target_file["builds"])

    build_type = the_build["buildRule"]
    if build_type != "executable":
        print(f"Error: {args.build} is not executable")
        exit(-1)
    name = the_build["outputName"]

    build_dir = make_build_path(path)
    if target_file["compilerFrontend"] == "msvc":
        command = build_dir / name
    else:
        command = build_dir / args.build / name
    forward_args = args.args if args.args else []
    str_repr = " ".join(forward_args)
    print(f"Running: {command} {str_repr}")
    return_code = subprocess.run([command] + forward_args)
    sys.exit(return_code.returncode)


def run_init(demo_zip: zipfile.ZipFile, subdir_name):
    project_dir = Path().cwd()
    dirs = ["include", "src", "lib", "tests", "builds"]
    dirs = [project_dir / x for x in dirs]
    print("Creating directories...")
    for a_dir in dirs:
        print(f"\t{str(a_dir)}")
        a_dir.mkdir(exist_ok=True)

    assert demo_zip, "Zip file error"
    print("Initialising from demo project...")

    files = [filename for filename in demo_zip.namelist() if filename.startswith(subdir_name)]
    assert len(files) > 0, f"Failed to find files for sub dir {subdir_name}"

    for file_name in files:
        with demo_zip.open(file_name) as the_file:
            file_path = Path(file_name)
            relative_path = file_path.relative_to(subdir_name)
            sys.stdout.write(f"\tCreating {str(relative_path)} ...")
            if relative_path.exists():
                print("warning, file already exists.")
                continue
            print("okay")

            relative_path.parent.mkdir(parents=True, exist_ok=True)
            relative_path.touch()
            relative_path.write_bytes(the_file.read())


def generate_flat_ninja_file(target_dict, project_dir, build_dir, args):
    frontend = target_dict["compilerFrontend"]
    project_ninja = build_dir / "build.ninja"

    builds = target_dict["builds"]

    with project_ninja.open("w+") as project_fd:
        project_writer = Writer(project_fd)
        # project_writer.include(str(build_dir / "rules.ninja"))
        if frontend == "msvc":
            msvcbuilds.add_compile(project_writer)
            msvcbuilds.add_ar(project_writer)
            msvcbuilds.add_exe(project_writer)
            msvcbuilds.add_shared(project_writer)
        elif frontend == "osx":
            # builder = osxbuilds.OsxBuilds()
            assert False, "OSX frontend is currently not supported."
        else:
            gccbuilds.add_compile(project_writer)
            gccbuilds.add_ar(project_writer)
            gccbuilds.add_exe(project_writer)
            gccbuilds.add_shared(project_writer)

        for build_info in builds:
            current_build = build_info
            current_build["directory"] = project_dir
            current_build["build_dir"] = build_dir

            if frontend == "msvc":
                builder = msvcbuilds.run_build
            elif frontend == "osx":
                # builder = osxbuilds.OsxBuilds()
                assert False, "OSX frontend is currently not supported."
            elif frontend == "gcc":
                builder = gccbuilds.run_build
            else:
                assert False, f"Error: Unknown frontend {frontend}"

            builder(build_info, target_dict, project_writer, args)


def make_build_path(target_path: Path):
    target_path = Path(target_path)
    if target_path.is_absolute():
        build_dir = target_path
    else:
        build_dir = Path().cwd() / target_path

    return build_dir


def make_project_path(root_dir: Path, build_dir: Path):
    project_dir = build_dir / root_dir
    assert project_dir.exists(), f"{str(project_dir)} does not exist."
    return project_dir


def vdir(obj):
    import inspect

    results = []

    # dir only gets the name of the attribute ...
    for x in dir(obj):
        # Remove _ and __ vars, as these aren't apart of our schema and will cause the validator to fail.
        if x.startswith("_"):
            continue

        # ... so when we want to check for modules, we need to explicitly access them using getattr.
        #
        # Remove modules, because we don't want to force users to do the following:
        # import os as _os.
        # Because that would be annoying.
        at = getattr(obj, x)
        if inspect.ismodule(at):
            continue

        results.append(x)
    return results


def load_target_py_file(file_path: Path):
    import importlib.util
    spec = importlib.util.spec_from_file_location("target", str(file_path))
    target_module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(target_module)
    return target_module


def convert_target_module_to_dict(target_module):
    keys = vdir(target_module)
    the_dict = {k: getattr(target_module, k) for k in keys}
    return the_dict


def run_build(build_name, target_path, skip_ninja_regen, args):
    print("Running build...")

    build_dir = make_build_path(target_path)
    file_path = build_dir / "target.py"

    completed_path = (Path().cwd() / file_path).resolve()
    assert file_path.exists(), f"Error: Could not find target.py at {str(completed_path)}"

    target_module = load_target_py_file(file_path)
    target_dict = convert_target_module_to_dict(target_module)

    builds = target_dict["builds"]
    the_build = find_build(build_name, builds)

    project_dir = make_project_path(target_dict["projectRoot"], build_dir)

    try:
        target_schema(target_dict, project_dir)
    except RuntimeError as exception:
        print(f"Error: {exception.args[0]}")
        sys.exit(-1)

    if not skip_ninja_regen:
        print("Generating ninja files...")
        generate_flat_ninja_file(target_dict, project_dir, build_dir, args)
        COMPILE_COMMANDS = True
        if COMPILE_COMMANDS:
            with (build_dir.resolve() / "compile_commands.json").open("w+") as cc_json:
                command = ["ninja", "-C", str(build_dir.resolve()), "-t", "compdb"]
                subprocess.run(command, stdout=cc_json, check=True)

    return run_ninja(build_dir, the_build["name"])


def add_naming_convention(
        output_name: str,
        build_type: BuildTypes,
        static_convention_func,
        dynamic_convention_func,
):
    if build_type == BuildTypes.staticLibrary:
        new_name = static_convention_func(output_name)
    elif build_type == BuildTypes.dynamicLibrary:
        new_name = dynamic_convention_func(output_name)
    else:
        assert False, f"Error Invalid build type: {build_type}"

    return new_name


def load_target_file(target_path):
    build_dir = make_build_path(target_path)

    target_file = build_dir / "target.py"
    assert target_file.exists(), f"Error: {str(target_file)} does not exists."

    target_module = load_target_py_file(target_file)
    target_dict = convert_target_module_to_dict(target_module)
    return target_dict

    # builds = target_dict["builds"]
    # return builds


def run_list(target_path):
    target_dict = load_target_file(target_path)
    builds = target_dict["builds"]

    frontend = target_dict["compilerFrontend"]

    if frontend == "msvc":
        static_convention_func = (
            msvcbuilds.windows_add_static_library_naming_convention
        )
        dynamic_convection_func = (
            msvcbuilds.windows_add_dynamic_library_naming_convention
        )
    elif frontend == "osx":
        assert False, "OSX frontend is currently not supported."
    elif frontend == "gcc":
        static_convention_func = (
            gccbuilds.linux_add_static_library_naming_convention
        )
        dynamic_convection_func = (
            gccbuilds.linux_add_dynamic_library_naming_convention
        )
    else:
        assert False, f"Error: Unknown compiler frontend: {frontend}"

    header = ["Item", "Name", "Build Rule", "Output Name"]
    table = []

    for number, build in enumerate(builds):
        build_type = BuildTypes[build["buildRule"]]
        if build_type in [BuildTypes.libraryReference, BuildTypes.headerOnly]:
            output_name = "n.a."
        elif build_type in [BuildTypes.staticLibrary, BuildTypes.dynamicLibrary]:
            output_name = add_naming_convention(
                build["outputName"],
                build_type,
                static_convention_func,
                dynamic_convection_func,
            )
        else:
            output_name = build["outputName"]

        row = [number, build["name"], build["buildRule"], output_name]
        table.append(row)

    print()
    print(tabulate(table, header))
    print()


def run_clobber(target_path):
    build_dir = Path().cwd()

    if target_path:
        target_path = Path(target_path)
        if target_path.is_absolute():
            build_dir = target_path
        else:
            build_dir = build_dir / Path(target_path)

    assert (build_dir / "target.py").exists(), (
        f"Failed to find target.py file in {str(build_dir)}.\n"
        "You might be trying to delete a directory that you want to keep."
    )

    print(f"Clobbering {str(build_dir)}...")

    dir_contents = build_dir.glob("*")
    for item in dir_contents:
        if item.name not in ["target.toml", "target.py"]:
            print(f"Deleting {item.name}")
            if item.is_dir():
                shutil.rmtree(str(item))
            else:
                os.remove(str(item))


if __name__ == "__main__":
    entry()
